import inspect
import typing

import faker
import pydantic
import pytest
import stamina
from polyfactory.factories.pydantic_factory import ModelFactory

import any_llm_client
from any_llm_client.clients.openai import ChatCompletionsRequest
from any_llm_client.clients.yandexgpt import YandexGPTRequest
from tests.conftest import LLMFuncRequest


def test_request_retry_config_default_kwargs_match() -> None:
    config_defaults: typing.Final = inspect.getfullargspec(any_llm_client.RequestRetryConfig).kwonlydefaults
    assert config_defaults
    stamina_defaults: typing.Final = inspect.getfullargspec(stamina.retry).kwonlydefaults
    assert stamina_defaults

    for one_ignored_setting in ("attempts",):
        config_defaults.pop(one_ignored_setting)
        stamina_defaults.pop(one_ignored_setting)

    assert config_defaults == stamina_defaults


def test_llm_error_str(faker: faker.Faker) -> None:
    response_content: typing.Final = faker.pystr().encode()
    assert str(any_llm_client.LLMError(response_content=response_content)) == f"(response_content={response_content!r})"


def test_llm_func_request_has_same_annotations_as_llm_client_methods() -> None:
    all_objects: typing.Final = (
        any_llm_client.LLMClient.request_llm_message,
        any_llm_client.LLMClient.stream_llm_message_chunks,
        LLMFuncRequest,
    )
    all_annotations: typing.Final = [typing.get_type_hints(one_object) for one_object in all_objects]

    for one_ignored_prop in ("return",):
        for annotations in all_annotations:
            if one_ignored_prop in annotations:
                annotations.pop(one_ignored_prop)

    assert all(annotations == all_annotations[0] for annotations in all_annotations)


@pytest.mark.parametrize("model_type", [YandexGPTRequest, ChatCompletionsRequest])
def test_dumped_llm_request_payload_dump_has_extra_data(model_type: type[pydantic.BaseModel]) -> None:
    extra: typing.Final = {"hi": "there", "hi-hi": "there-there"}
    generated_data: typing.Final = ModelFactory.create_factory(model_type).build(**extra).model_dump(by_alias=True)  # type: ignore[arg-type]
    dumped_model: typing.Final = model_type(**{**generated_data, **extra}).model_dump(mode="json", by_alias=True)

    assert dumped_model["hi"] == "there"
    assert dumped_model["hi-hi"] == "there-there"



import contextlib
import dataclasses
import os
import types
import typing
from http import HTTPStatus

import annotated_types
import httpx
import pydantic
import typing_extensions

from any_llm_client.core import LLMClient, LLMConfig, LLMError, Message, OutOfTokensOrSymbolsError, UserMessage
from any_llm_client.http import get_http_client_from_kwargs, make_http_request, make_streaming_http_request
from any_llm_client.retry import RequestRetryConfig


YANDEXGPT_AUTH_HEADER_ENV_NAME: typing.Final = "ANY_LLM_CLIENT_YANDEXGPT_AUTH_HEADER"
YANDEXGPT_FOLDER_ID_ENV_NAME: typing.Final = "ANY_LLM_CLIENT_YANDEXGPT_FOLDER_ID"


class YandexGPTConfig(LLMConfig):
    if typing.TYPE_CHECKING:
        url: str = "https://llm.api.cloud.yandex.net/foundationModels/v1/completion"
    else:
        url: pydantic.HttpUrl = "https://llm.api.cloud.yandex.net/foundationModels/v1/completion"
    auth_header: str = pydantic.Field(  # type: ignore[assignment]
        default_factory=lambda: os.environ.get(YANDEXGPT_AUTH_HEADER_ENV_NAME), validate_default=True
    )
    folder_id: str = pydantic.Field(  # type: ignore[assignment]
        default_factory=lambda: os.environ.get(YANDEXGPT_FOLDER_ID_ENV_NAME), validate_default=True
    )
    model_name: str
    model_version: str = "latest"
    max_tokens: int = 7400
    api_type: typing.Literal["yandexgpt"] = "yandexgpt"


class YandexGPTCompletionOptions(pydantic.BaseModel):
    stream: bool
    temperature: float = 0.2
    max_tokens: int = pydantic.Field(gt=0, alias="maxTokens")


class YandexGPTRequest(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(protected_namespaces=(), extra="allow")
    model_uri: str = pydantic.Field(alias="modelUri")
    completion_options: YandexGPTCompletionOptions = pydantic.Field(alias="completionOptions")
    messages: list[Message]


class YandexGPTAlternative(pydantic.BaseModel):
    message: Message


class YandexGPTResult(pydantic.BaseModel):
    alternatives: typing.Annotated[list[YandexGPTAlternative], annotated_types.MinLen(1)]


class YandexGPTResponse(pydantic.BaseModel):
    result: YandexGPTResult


def _handle_status_error(*, status_code: int, content: bytes) -> typing.NoReturn:
    if status_code == HTTPStatus.BAD_REQUEST and (
        b"number of input tokens must be no more than" in content
        or (b"text length is" in content and b"which is outside the range" in content)
    ):
        raise OutOfTokensOrSymbolsError(response_content=content)
    raise LLMError(response_content=content)


@dataclasses.dataclass(slots=True, init=False)
class YandexGPTClient(LLMClient):
    config: YandexGPTConfig
    httpx_client: httpx.AsyncClient
    request_retry: RequestRetryConfig

    def __init__(
        self,
        config: YandexGPTConfig,
        *,
        request_retry: RequestRetryConfig | None = None,
        **httpx_kwargs: typing.Any,  # noqa: ANN401
    ) -> None:
        self.config = config
        self.request_retry = request_retry or RequestRetryConfig()
        self.httpx_client = get_http_client_from_kwargs(httpx_kwargs)

    def _build_request(self, payload: dict[str, typing.Any]) -> httpx.Request:
        return self.httpx_client.build_request(
            method="POST",
            url=str(self.config.url),
            json=payload,
            headers={"Authorization": self.config.auth_header, "x-data-logging-enabled": "false"},
        )

    def _prepare_payload(
        self,
        *,
        messages: str | list[Message],
        temperature: float = 0.2,
        stream: bool,
        extra: dict[str, typing.Any] | None,
    ) -> dict[str, typing.Any]:
        messages = [UserMessage(messages)] if isinstance(messages, str) else messages
        return YandexGPTRequest(
            modelUri=f"gpt://{self.config.folder_id}/{self.config.model_name}/{self.config.model_version}",
            completionOptions=YandexGPTCompletionOptions(
                stream=stream, temperature=temperature, maxTokens=self.config.max_tokens
            ),
            messages=messages,
            **extra or {},
        ).model_dump(mode="json", by_alias=True)

    async def request_llm_message(
        self, messages: str | list[Message], *, temperature: float = 0.2, extra: dict[str, typing.Any] | None = None
    ) -> str:
        payload: typing.Final = self._prepare_payload(
            messages=messages, temperature=temperature, stream=False, extra=extra
        )

        try:
            response: typing.Final = await make_http_request(
                httpx_client=self.httpx_client,
                request_retry=self.request_retry,
                build_request=lambda: self._build_request(payload),
            )
        except httpx.HTTPStatusError as exception:
            _handle_status_error(status_code=exception.response.status_code, content=exception.response.content)

        return YandexGPTResponse.model_validate_json(response.content).result.alternatives[0].message.text

    async def _iter_response_chunks(self, response: httpx.Response) -> typing.AsyncIterable[str]:
        previous_cursor = 0
        async for one_line in response.aiter_lines():
            validated_response = YandexGPTResponse.model_validate_json(one_line)
            response_text = validated_response.result.alternatives[0].message.text
            yield response_text[previous_cursor:]
            previous_cursor = len(response_text)

    @contextlib.asynccontextmanager
    async def stream_llm_message_chunks(
        self, messages: str | list[Message], *, temperature: float = 0.2, extra: dict[str, typing.Any] | None = None
    ) -> typing.AsyncIterator[typing.AsyncIterable[str]]:
        payload: typing.Final = self._prepare_payload(
            messages=messages, temperature=temperature, stream=True, extra=extra
        )

        try:
            async with make_streaming_http_request(
                httpx_client=self.httpx_client,
                request_retry=self.request_retry,
                build_request=lambda: self._build_request(payload),
            ) as response:
                yield self._iter_response_chunks(response)
        except httpx.HTTPStatusError as exception:
            content: typing.Final = await exception.response.aread()
            await exception.response.aclose()
            _handle_status_error(status_code=exception.response.status_code, content=content)

    async def __aenter__(self) -> typing_extensions.Self:
        await self.httpx_client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        await self.httpx_client.__aexit__(exc_type=exc_type, exc_value=exc_value, traceback=traceback)



import contextlib
import typing

import pytest
import stamina
from polyfactory.factories.typed_dict_factory import TypedDictFactory

import any_llm_client


@pytest.fixture(scope="session", autouse=True)
def anyio_backend() -> str:
    return "asyncio"


@pytest.fixture(scope="session", autouse=True)
def _deactivate_retries() -> None:
    stamina.set_active(False)


class LLMFuncRequest(typing.TypedDict):
    messages: str | list[any_llm_client.Message]
    temperature: float
    extra: dict[str, typing.Any] | None


class LLMFuncRequestFactory(TypedDictFactory[LLMFuncRequest]): ...


async def consume_llm_message_chunks(
    stream_llm_message_chunks_context_manager: contextlib._AsyncGeneratorContextManager[typing.AsyncIterable[str]],
    /,
) -> list[str]:
    async with stream_llm_message_chunks_context_manager as response_iterable:
        return [one_item async for one_item in response_iterable]



import asyncio
import os

import pydantic_settings

import any_llm_client


class Settings(pydantic_settings.BaseSettings):
    llm_model: any_llm_client.AnyLLMConfig


os.environ["LLM_MODEL"] = """{
    "api_type": "openai",
    "url": "http://127.0.0.1:11434/v1/chat/completions",
    "model_name": "qwen2.5-coder:1.5b"
}"""
settings = Settings()


async def main() -> None:
    async with any_llm_client.get_client(settings.llm_model) as client:
        print(await client.request_llm_message("Кек, чо как вообще на нарах?"))


asyncio.run(main())



import typing

import faker
import httpx
import pydantic
import pytest
from polyfactory.factories.pydantic_factory import ModelFactory

import any_llm_client
from any_llm_client.clients.yandexgpt import YandexGPTAlternative, YandexGPTResponse, YandexGPTResult
from tests.conftest import LLMFuncRequestFactory, consume_llm_message_chunks


class YandexGPTConfigFactory(ModelFactory[any_llm_client.YandexGPTConfig]): ...


class TestYandexGPTRequestLLMResponse:
    async def test_ok(self, faker: faker.Faker) -> None:
        expected_result: typing.Final = faker.pystr()
        response: typing.Final = httpx.Response(
            200,
            json=YandexGPTResponse(
                result=YandexGPTResult(
                    alternatives=[YandexGPTAlternative(message=any_llm_client.AssistantMessage(expected_result))]
                )
            ).model_dump(mode="json"),
        )

        result: typing.Final = await any_llm_client.get_client(
            YandexGPTConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        ).request_llm_message(**LLMFuncRequestFactory.build())

        assert result == expected_result

    async def test_fails_without_alternatives(self) -> None:
        response: typing.Final = httpx.Response(
            200, json=YandexGPTResponse(result=YandexGPTResult.model_construct(alternatives=[])).model_dump(mode="json")
        )
        client: typing.Final = any_llm_client.get_client(
            YandexGPTConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        )

        with pytest.raises(pydantic.ValidationError):
            await client.request_llm_message(**LLMFuncRequestFactory.build())


class TestYandexGPTRequestLLMMessageChunks:
    async def test_ok(self, faker: faker.Faker) -> None:
        expected_result: typing.Final = faker.pylist(value_types=[str])
        config: typing.Final = YandexGPTConfigFactory.build()
        func_request: typing.Final = LLMFuncRequestFactory.build()
        response_content: typing.Final = (
            "\n".join(
                YandexGPTResponse(
                    result=YandexGPTResult(
                        alternatives=[
                            YandexGPTAlternative(
                                message=any_llm_client.AssistantMessage("".join(expected_result[: one_index + 1]))
                            )
                        ]
                    )
                ).model_dump_json()
                for one_index in range(len(expected_result))
            )
            + "\n"
        )
        response: typing.Final = httpx.Response(200, content=response_content)

        result: typing.Final = await consume_llm_message_chunks(
            any_llm_client.get_client(
                config, transport=httpx.MockTransport(lambda _: response)
            ).stream_llm_message_chunks(**func_request)
        )

        assert result == expected_result

    async def test_fails_without_alternatives(self) -> None:
        response_content: typing.Final = (
            YandexGPTResponse(result=YandexGPTResult.model_construct(alternatives=[])).model_dump_json() + "\n"
        )
        response: typing.Final = httpx.Response(200, content=response_content)

        client: typing.Final = any_llm_client.get_client(
            YandexGPTConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        )

        with pytest.raises(pydantic.ValidationError):
            await consume_llm_message_chunks(client.stream_llm_message_chunks(**LLMFuncRequestFactory.build()))


class TestYandexGPTLLMErrors:
    @pytest.mark.parametrize("stream", [True, False])
    @pytest.mark.parametrize("status_code", [400, 500])
    async def test_fails_with_unknown_error(self, stream: bool, status_code: int) -> None:
        client: typing.Final = any_llm_client.get_client(
            YandexGPTConfigFactory.build(), transport=httpx.MockTransport(lambda _: httpx.Response(status_code))
        )

        coroutine: typing.Final = (
            consume_llm_message_chunks(client.stream_llm_message_chunks(**LLMFuncRequestFactory.build()))
            if stream
            else client.request_llm_message(**LLMFuncRequestFactory.build())
        )

        with pytest.raises(any_llm_client.LLMError) as exc_info:
            await coroutine
        assert type(exc_info.value) is any_llm_client.LLMError

    @pytest.mark.parametrize("stream", [True, False])
    @pytest.mark.parametrize(
        "response_content",
        [
            b"...folder_id=1111: number of input tokens must be no more than 8192, got 28498...",
            b"...folder_id=1111: text length is 349354, which is outside the range (0, 100000]...",
        ],
    )
    async def test_fails_with_out_of_tokens_error(self, stream: bool, response_content: bytes | None) -> None:
        response: typing.Final = httpx.Response(400, content=response_content)
        client: typing.Final = any_llm_client.get_client(
            YandexGPTConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        )

        coroutine: typing.Final = (
            consume_llm_message_chunks(client.stream_llm_message_chunks(**LLMFuncRequestFactory.build()))
            if stream
            else client.request_llm_message(**LLMFuncRequestFactory.build())
        )

        with pytest.raises(any_llm_client.OutOfTokensOrSymbolsError):
            await coroutine



import typing

from polyfactory.factories.pydantic_factory import ModelFactory

import any_llm_client
from tests.conftest import LLMFuncRequestFactory, consume_llm_message_chunks


class MockLLMConfigFactory(ModelFactory[any_llm_client.MockLLMConfig]): ...


async def test_mock_client_request_llm_message_returns_config_value() -> None:
    config: typing.Final = MockLLMConfigFactory.build()
    response: typing.Final = await any_llm_client.get_client(config).request_llm_message(
        **LLMFuncRequestFactory.build()
    )
    assert response == config.response_message


async def test_mock_client_stream_llm_message_chunks_returns_config_value() -> None:
    config: typing.Final = MockLLMConfigFactory.build()
    response: typing.Final = await consume_llm_message_chunks(
        any_llm_client.get_client(config).stream_llm_message_chunks(**LLMFuncRequestFactory.build())
    )
    assert response == config.stream_messages



import contextlib
import dataclasses
import enum
import types
import typing

import pydantic
import typing_extensions


class MessageRole(str, enum.Enum):
    system = "system"
    user = "user"
    assistant = "assistant"


@pydantic.dataclasses.dataclass(kw_only=True)
class Message:
    role: MessageRole
    text: str


if typing.TYPE_CHECKING:

    @pydantic.dataclasses.dataclass
    class SystemMessage(Message):
        role: typing.Literal[MessageRole.system] = pydantic.Field(MessageRole.system, init=False)
        text: str

    @pydantic.dataclasses.dataclass
    class UserMessage(Message):
        role: typing.Literal[MessageRole.user] = pydantic.Field(MessageRole.user, init=False)
        text: str

    @pydantic.dataclasses.dataclass
    class AssistantMessage(Message):
        role: typing.Literal[MessageRole.assistant] = pydantic.Field(MessageRole.assistant, init=False)
        text: str
else:

    def SystemMessage(text: str) -> Message:  # noqa: N802
        return Message(role=MessageRole.system, text=text)

    def UserMessage(text: str) -> Message:  # noqa: N802
        return Message(role=MessageRole.user, text=text)

    def AssistantMessage(text: str) -> Message:  # noqa: N802
        return Message(role=MessageRole.assistant, text=text)


class LLMConfig(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(protected_namespaces=())
    api_type: str


@dataclasses.dataclass(slots=True, init=False)
class LLMClient(typing.Protocol):
    async def request_llm_message(
        self, messages: str | list[Message], *, temperature: float = 0.2, extra: dict[str, typing.Any] | None = None
    ) -> str: ...  # raises LLMError

    @contextlib.asynccontextmanager
    def stream_llm_message_chunks(
        self, messages: str | list[Message], *, temperature: float = 0.2, extra: dict[str, typing.Any] | None = None
    ) -> typing.AsyncIterator[typing.AsyncIterable[str]]: ...  # raises LLMError

    async def __aenter__(self) -> typing_extensions.Self: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None: ...


@dataclasses.dataclass
class LLMError(Exception):
    response_content: bytes

    def __str__(self) -> str:
        return self.__repr__().removeprefix(self.__class__.__name__)


@dataclasses.dataclass
class OutOfTokensOrSymbolsError(LLMError): ...



import typing

import faker
import httpx
import pydantic
import pytest
from polyfactory.factories.pydantic_factory import ModelFactory

import any_llm_client
from any_llm_client.clients.openai import (
    ChatCompletionsMessage,
    ChatCompletionsNotStreamingResponse,
    ChatCompletionsStreamingEvent,
    OneNotStreamingChoice,
    OneStreamingChoice,
    OneStreamingChoiceDelta,
)
from tests.conftest import LLMFuncRequestFactory, consume_llm_message_chunks


class OpenAIConfigFactory(ModelFactory[any_llm_client.OpenAIConfig]): ...


class TestOpenAIRequestLLMResponse:
    async def test_ok(self, faker: faker.Faker) -> None:
        expected_result: typing.Final = faker.pystr()
        response: typing.Final = httpx.Response(
            200,
            json=ChatCompletionsNotStreamingResponse(
                choices=[
                    OneNotStreamingChoice(
                        message=ChatCompletionsMessage(
                            role=any_llm_client.MessageRole.assistant, content=expected_result
                        )
                    )
                ]
            ).model_dump(mode="json"),
        )

        result: typing.Final = await any_llm_client.get_client(
            OpenAIConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        ).request_llm_message(**LLMFuncRequestFactory.build())

        assert result == expected_result

    async def test_fails_without_alternatives(self) -> None:
        response: typing.Final = httpx.Response(
            200,
            json=ChatCompletionsNotStreamingResponse.model_construct(choices=[]).model_dump(mode="json"),
        )
        client: typing.Final = any_llm_client.get_client(
            OpenAIConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        )

        with pytest.raises(pydantic.ValidationError):
            await client.request_llm_message(**LLMFuncRequestFactory.build())


class TestOpenAIRequestLLMMessageChunks:
    async def test_ok(self, faker: faker.Faker) -> None:
        generated_messages: typing.Final = [
            OneStreamingChoiceDelta(role=any_llm_client.MessageRole.assistant),
            OneStreamingChoiceDelta(content="H"),
            OneStreamingChoiceDelta(content="i"),
            OneStreamingChoiceDelta(content=" t"),
            OneStreamingChoiceDelta(role=any_llm_client.MessageRole.assistant, content="here"),
            OneStreamingChoiceDelta(),
            OneStreamingChoiceDelta(content=". How is you"),
            OneStreamingChoiceDelta(content="r day?"),
            OneStreamingChoiceDelta(),
        ]
        expected_result: typing.Final = [
            "H",
            "i",
            " t",
            "here",
            ". How is you",
            "r day?",
        ]
        config: typing.Final = OpenAIConfigFactory.build()
        func_request: typing.Final = LLMFuncRequestFactory.build()
        response_content: typing.Final = (
            "\n\n".join(
                "data: "
                + ChatCompletionsStreamingEvent(choices=[OneStreamingChoice(delta=one_message)]).model_dump_json()
                for one_message in generated_messages
            )
            + f"\n\ndata: [DONE]\n\ndata: {faker.pystr()}\n\n"
        )
        response: typing.Final = httpx.Response(
            200, headers={"Content-Type": "text/event-stream"}, content=response_content
        )
        client: typing.Final = any_llm_client.get_client(config, transport=httpx.MockTransport(lambda _: response))

        result: typing.Final = await consume_llm_message_chunks(client.stream_llm_message_chunks(**func_request))

        assert result == expected_result

    async def test_fails_without_alternatives(self) -> None:
        response_content: typing.Final = (
            f"data: {ChatCompletionsStreamingEvent.model_construct(choices=[]).model_dump_json()}\n\n"
        )
        response: typing.Final = httpx.Response(
            200, headers={"Content-Type": "text/event-stream"}, content=response_content
        )
        client: typing.Final = any_llm_client.get_client(
            OpenAIConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        )

        with pytest.raises(pydantic.ValidationError):
            await consume_llm_message_chunks(client.stream_llm_message_chunks(**LLMFuncRequestFactory.build()))


class TestOpenAILLMErrors:
    @pytest.mark.parametrize("stream", [True, False])
    @pytest.mark.parametrize("status_code", [400, 500])
    async def test_fails_with_unknown_error(self, stream: bool, status_code: int) -> None:
        client: typing.Final = any_llm_client.get_client(
            OpenAIConfigFactory.build(), transport=httpx.MockTransport(lambda _: httpx.Response(status_code))
        )

        coroutine: typing.Final = (
            consume_llm_message_chunks(client.stream_llm_message_chunks(**LLMFuncRequestFactory.build()))
            if stream
            else client.request_llm_message(**LLMFuncRequestFactory.build())
        )

        with pytest.raises(any_llm_client.LLMError) as exc_info:
            await coroutine
        assert type(exc_info.value) is any_llm_client.LLMError

    @pytest.mark.parametrize("stream", [True, False])
    @pytest.mark.parametrize(
        "content",
        [
            b'{"object":"error","message":"This model\'s maximum context length is 4096 tokens. However, you requested 5253 tokens in the messages, Please reduce the length of the messages.","type":"BadRequestError","param":null,"code":400}',  # noqa: E501
            b'{"object":"error","message":"This model\'s maximum context length is 16384 tokens. However, you requested 100000 tokens in the messages, Please reduce the length of the messages.","type":"BadRequestError","param":null,"code":400}',  # noqa: E501
        ],
    )
    async def test_fails_with_out_of_tokens_error(self, stream: bool, content: bytes | None) -> None:
        response: typing.Final = httpx.Response(400, content=content)
        client: typing.Final = any_llm_client.get_client(
            OpenAIConfigFactory.build(), transport=httpx.MockTransport(lambda _: response)
        )

        coroutine: typing.Final = (
            consume_llm_message_chunks(client.stream_llm_message_chunks(**LLMFuncRequestFactory.build()))
            if stream
            else client.request_llm_message(**LLMFuncRequestFactory.build())
        )

        with pytest.raises(any_llm_client.OutOfTokensOrSymbolsError):
            await coroutine


class TestOpenAIMessageAlternation:
    @pytest.mark.parametrize(
        ("messages", "expected_result"),
        [
            ([], []),
            ([any_llm_client.SystemMessage("")], []),
            ([any_llm_client.SystemMessage(" ")], []),
            ([any_llm_client.UserMessage("")], []),
            ([any_llm_client.AssistantMessage("")], []),
            ([any_llm_client.SystemMessage(""), any_llm_client.UserMessage("")], []),
            ([any_llm_client.SystemMessage(""), any_llm_client.AssistantMessage("")], []),
            (
                [
                    any_llm_client.SystemMessage(""),
                    any_llm_client.UserMessage(""),
                    any_llm_client.AssistantMessage(""),
                    any_llm_client.AssistantMessage(""),
                    any_llm_client.UserMessage(""),
                    any_llm_client.AssistantMessage(""),
                ],
                [],
            ),
            (
                [any_llm_client.SystemMessage("Be nice")],
                [ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Be nice")],
            ),
            (
                [any_llm_client.UserMessage("Hi there"), any_llm_client.AssistantMessage("Hi! How can I help you?")],
                [
                    ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Hi there"),
                    ChatCompletionsMessage(
                        role=any_llm_client.MessageRole.assistant, content="Hi! How can I help you?"
                    ),
                ],
            ),
            (
                [
                    any_llm_client.SystemMessage(""),
                    any_llm_client.UserMessage("Hi there"),
                    any_llm_client.AssistantMessage("Hi! How can I help you?"),
                ],
                [
                    ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Hi there"),
                    ChatCompletionsMessage(
                        role=any_llm_client.MessageRole.assistant, content="Hi! How can I help you?"
                    ),
                ],
            ),
            (
                [any_llm_client.SystemMessage("Be nice"), any_llm_client.UserMessage("Hi there")],
                [ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Be nice\n\nHi there")],
            ),
            (
                [
                    any_llm_client.SystemMessage("Be nice"),
                    any_llm_client.AssistantMessage("Hi!"),
                    any_llm_client.AssistantMessage("I'm your answer to everything."),
                    any_llm_client.AssistantMessage("How can I help you?"),
                    any_llm_client.UserMessage("Hi there"),
                    any_llm_client.UserMessage(""),
                    any_llm_client.UserMessage("Why is the sky blue?"),
                    any_llm_client.AssistantMessage(" "),
                    any_llm_client.AssistantMessage("Well..."),
                    any_llm_client.AssistantMessage(""),
                    any_llm_client.AssistantMessage(" \n "),
                    any_llm_client.UserMessage("Hmmm..."),
                ],
                [
                    ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Be nice"),
                    ChatCompletionsMessage(
                        role=any_llm_client.MessageRole.assistant,
                        content="Hi!\n\nI'm your answer to everything.\n\nHow can I help you?",
                    ),
                    ChatCompletionsMessage(
                        role=any_llm_client.MessageRole.user, content="Hi there\n\nWhy is the sky blue?"
                    ),
                    ChatCompletionsMessage(role=any_llm_client.MessageRole.assistant, content="Well..."),
                    ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Hmmm..."),
                ],
            ),
        ],
    )
    def test_with_alternation(
        self, messages: list[any_llm_client.Message], expected_result: list[ChatCompletionsMessage]
    ) -> None:
        client: typing.Final = any_llm_client.OpenAIClient(
            OpenAIConfigFactory.build(force_user_assistant_message_alternation=True)
        )
        assert client._prepare_messages(messages) == expected_result  # noqa: SLF001

    def test_without_alternation(self) -> None:
        client: typing.Final = any_llm_client.OpenAIClient(
            OpenAIConfigFactory.build(force_user_assistant_message_alternation=False)
        )
        assert client._prepare_messages(  # noqa: SLF001
            [any_llm_client.SystemMessage("Be nice"), any_llm_client.UserMessage("Hi there")]
        ) == [
            ChatCompletionsMessage(role=any_llm_client.MessageRole.system, content="Be nice"),
            ChatCompletionsMessage(role=any_llm_client.MessageRole.user, content="Hi there"),
        ]



import pydantic
import pytest
from polyfactory.factories.pydantic_factory import ModelFactory

import any_llm_client


class ConfigHolder(pydantic.BaseModel):
    config: any_llm_client.AnyLLMConfig


@pytest.mark.parametrize(
    "config", [one_holder.config for one_holder in ModelFactory[ConfigHolder].create_factory(ConfigHolder).coverage()]
)
async def test_lifespan(config: any_llm_client.AnyLLMConfig) -> None:
    async with any_llm_client.get_client(config):
        pass



import asyncio
import os

import any_llm_client


config = any_llm_client.YandexGPTConfig(
    auth_header=os.environ["YANDEX_AUTH_HEADER"], folder_id=os.environ["YANDEX_FOLDER_ID"], model_name="yandexgpt"
)


async def main() -> None:
    async with any_llm_client.get_client(config) as client:
        print(await client.request_llm_message("Кек, чо как вообще на нарах?"))


asyncio.run(main())



import dataclasses
import datetime


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class RequestRetryConfig:
    """Request retry configuration that is passed to `stamina.retry`. Applies to httpx.HTTPError.

    Uses defaults from `stamina.retry` except for attempts: by default 3 instead of 10.
    See more at https://stamina.hynek.me/en/stable/api.html#stamina.retry
    """

    attempts: int | None = 3
    "Maximum total number of attempts. Can be combined with *timeout*."
    timeout: float | datetime.timedelta | None = 45.0
    "Maximum total time for all retries. Can be combined with *attempts*."
    wait_initial: float | datetime.timedelta = 0.1
    "Minimum backoff before the *first* retry."
    wait_max: float | datetime.timedelta = 5.0
    "Maximum backoff time between retries at any time."
    wait_jitter: float | datetime.timedelta = 1.0
    "Maximum *jitter* that is added to retry back-off delays (the actual jitter added is a random number between 0 and *wait_jitter*)"  # noqa: E501
    wait_exp_base: float = 2.0
    "The exponential base used to compute the retry backoff."



import faker
import pytest

import any_llm_client


def test_unknown_client_raises_assertion_error(faker: faker.Faker) -> None:
    with pytest.raises(AssertionError):
        any_llm_client.get_client(faker.pyobject())  # type: ignore[arg-type]



import contextlib
import dataclasses
import os
import types
import typing
from http import HTTPStatus

import annotated_types
import httpx
import httpx_sse
import pydantic
import typing_extensions

from any_llm_client.core import (
    LLMClient,
    LLMConfig,
    LLMError,
    Message,
    MessageRole,
    OutOfTokensOrSymbolsError,
    UserMessage,
)
from any_llm_client.http import get_http_client_from_kwargs, make_http_request, make_streaming_http_request
from any_llm_client.retry import RequestRetryConfig


OPENAI_AUTH_TOKEN_ENV_NAME: typing.Final = "ANY_LLM_CLIENT_OPENAI_AUTH_TOKEN"  # noqa: S105


class OpenAIConfig(LLMConfig):
    if typing.TYPE_CHECKING:
        url: str
    else:
        url: pydantic.HttpUrl
    auth_token: str | None = pydantic.Field(default_factory=lambda: os.environ.get(OPENAI_AUTH_TOKEN_ENV_NAME))
    model_name: str
    request_extra: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    force_user_assistant_message_alternation: bool = False
    "Gemma 2 doesn't support {role: system, text: ...} message, and requires alternated messages"
    api_type: typing.Literal["openai"] = "openai"


class ChatCompletionsMessage(pydantic.BaseModel):
    role: MessageRole
    content: str


class ChatCompletionsRequest(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    stream: bool
    model: str
    messages: list[ChatCompletionsMessage]
    temperature: float = 0.2


class OneStreamingChoiceDelta(pydantic.BaseModel):
    role: typing.Literal[MessageRole.assistant] | None = None
    content: str | None = None


class OneStreamingChoice(pydantic.BaseModel):
    delta: OneStreamingChoiceDelta


class ChatCompletionsStreamingEvent(pydantic.BaseModel):
    choices: typing.Annotated[list[OneStreamingChoice], annotated_types.MinLen(1)]


class OneNotStreamingChoice(pydantic.BaseModel):
    message: ChatCompletionsMessage


class ChatCompletionsNotStreamingResponse(pydantic.BaseModel):
    choices: typing.Annotated[list[OneNotStreamingChoice], annotated_types.MinLen(1)]


def _make_user_assistant_alternate_messages(
    messages: typing.Iterable[ChatCompletionsMessage],
) -> typing.Iterable[ChatCompletionsMessage]:
    current_message_role = MessageRole.user
    current_message_content_chunks = []

    for one_message in messages:
        if not one_message.content.strip():
            continue

        if (
            one_message.role in {MessageRole.system, MessageRole.user} and current_message_role == MessageRole.user
        ) or one_message.role == current_message_role == MessageRole.assistant:
            current_message_content_chunks.append(one_message.content)
        else:
            if current_message_content_chunks:
                yield ChatCompletionsMessage(
                    role=current_message_role, content="\n\n".join(current_message_content_chunks)
                )
            current_message_content_chunks = [one_message.content]
            current_message_role = one_message.role

    if current_message_content_chunks:
        yield ChatCompletionsMessage(role=current_message_role, content="\n\n".join(current_message_content_chunks))


def _handle_status_error(*, status_code: int, content: bytes) -> typing.NoReturn:
    if status_code == HTTPStatus.BAD_REQUEST and b"Please reduce the length of the messages" in content:  # vLLM
        raise OutOfTokensOrSymbolsError(response_content=content)
    raise LLMError(response_content=content)


@dataclasses.dataclass(slots=True, init=False)
class OpenAIClient(LLMClient):
    config: OpenAIConfig
    httpx_client: httpx.AsyncClient
    request_retry: RequestRetryConfig

    def __init__(
        self,
        config: OpenAIConfig,
        *,
        request_retry: RequestRetryConfig | None = None,
        **httpx_kwargs: typing.Any,  # noqa: ANN401
    ) -> None:
        self.config = config
        self.request_retry = request_retry or RequestRetryConfig()
        self.httpx_client = get_http_client_from_kwargs(httpx_kwargs)

    def _build_request(self, payload: dict[str, typing.Any]) -> httpx.Request:
        return self.httpx_client.build_request(
            method="POST",
            url=str(self.config.url),
            json=payload,
            headers={"Authorization": f"Bearer {self.config.auth_token}"} if self.config.auth_token else None,
        )

    def _prepare_messages(self, messages: str | list[Message]) -> list[ChatCompletionsMessage]:
        messages = [UserMessage(messages)] if isinstance(messages, str) else messages
        initial_messages: typing.Final = (
            ChatCompletionsMessage(role=one_message.role, content=one_message.text) for one_message in messages
        )
        return (
            list(_make_user_assistant_alternate_messages(initial_messages))
            if self.config.force_user_assistant_message_alternation
            else list(initial_messages)
        )

    async def request_llm_message(
        self, messages: str | list[Message], *, temperature: float = 0.2, extra: dict[str, typing.Any] | None = None
    ) -> str:
        payload: typing.Final = ChatCompletionsRequest(
            stream=False,
            model=self.config.model_name,
            messages=self._prepare_messages(messages),
            temperature=temperature,
            **self.config.request_extra | (extra or {}),
        ).model_dump(mode="json")
        try:
            response: typing.Final = await make_http_request(
                httpx_client=self.httpx_client,
                request_retry=self.request_retry,
                build_request=lambda: self._build_request(payload),
            )
        except httpx.HTTPStatusError as exception:
            _handle_status_error(status_code=exception.response.status_code, content=exception.response.content)
        try:
            return ChatCompletionsNotStreamingResponse.model_validate_json(response.content).choices[0].message.content
        finally:
            await response.aclose()

    async def _iter_response_chunks(self, response: httpx.Response) -> typing.AsyncIterable[str]:
        async for event in httpx_sse.EventSource(response).aiter_sse():
            if event.data == "[DONE]":
                break
            validated_response = ChatCompletionsStreamingEvent.model_validate_json(event.data)
            if not (one_chunk := validated_response.choices[0].delta.content):
                continue
            yield one_chunk

    @contextlib.asynccontextmanager
    async def stream_llm_message_chunks(
        self, messages: str | list[Message], *, temperature: float = 0.2, extra: dict[str, typing.Any] | None = None
    ) -> typing.AsyncIterator[typing.AsyncIterable[str]]:
        payload: typing.Final = ChatCompletionsRequest(
            stream=True,
            model=self.config.model_name,
            messages=self._prepare_messages(messages),
            temperature=temperature,
            **self.config.request_extra | (extra or {}),
        ).model_dump(mode="json")
        try:
            async with make_streaming_http_request(
                httpx_client=self.httpx_client,
                request_retry=self.request_retry,
                build_request=lambda: self._build_request(payload),
            ) as response:
                yield self._iter_response_chunks(response)
        except httpx.HTTPStatusError as exception:
            content: typing.Final = await exception.response.aread()
            await exception.response.aclose()
            _handle_status_error(status_code=exception.response.status_code, content=content)

    async def __aenter__(self) -> typing_extensions.Self:
        await self.httpx_client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        await self.httpx_client.__aexit__(exc_type=exc_type, exc_value=exc_value, traceback=traceback)



import functools
import typing

import pydantic

from any_llm_client.clients.mock import MockLLMClient, MockLLMConfig
from any_llm_client.clients.openai import OpenAIClient, OpenAIConfig
from any_llm_client.clients.yandexgpt import YandexGPTClient, YandexGPTConfig
from any_llm_client.core import LLMClient
from any_llm_client.retry import RequestRetryConfig


AnyLLMConfig = typing.Annotated[YandexGPTConfig | OpenAIConfig | MockLLMConfig, pydantic.Discriminator("api_type")]


if typing.TYPE_CHECKING:

    def get_client(
        config: AnyLLMConfig,
        *,
        request_retry: RequestRetryConfig | None = None,
        **httpx_kwargs: typing.Any,  # noqa: ANN401
    ) -> LLMClient: ...
else:

    @functools.singledispatch
    def get_client(
        config: typing.Any,  # noqa: ANN401, ARG001
        *,
        request_retry: RequestRetryConfig | None = None,  # noqa: ARG001
        **httpx_kwargs: typing.Any,  # noqa: ANN401, ARG001
    ) -> LLMClient:
        raise AssertionError("unknown LLM config type")

    @get_client.register
    def _(
        config: YandexGPTConfig,
        *,
        request_retry: RequestRetryConfig | None = None,
        **httpx_kwargs: typing.Any,  # noqa: ANN401
    ) -> LLMClient:
        return YandexGPTClient(config=config, request_retry=request_retry, **httpx_kwargs)

    @get_client.register
    def _(
        config: OpenAIConfig,
        *,
        request_retry: RequestRetryConfig | None = None,
        **httpx_kwargs: typing.Any,  # noqa: ANN401
    ) -> LLMClient:
        return OpenAIClient(config=config, request_retry=request_retry, **httpx_kwargs)

    @get_client.register
    def _(
        config: MockLLMConfig,
        *,
        request_retry: RequestRetryConfig | None = None,  # noqa: ARG001
        **httpx_kwargs: typing.Any,  # noqa: ANN401, ARG001
    ) -> LLMClient:
        return MockLLMClient(config=config)



from any_llm_client.clients.mock import MockLLMClient, MockLLMConfig
from any_llm_client.clients.openai import OpenAIClient, OpenAIConfig
from any_llm_client.clients.yandexgpt import YandexGPTClient, YandexGPTConfig
from any_llm_client.core import (
    AssistantMessage,
    LLMClient,
    LLMConfig,
    LLMError,
    Message,
    MessageRole,
    OutOfTokensOrSymbolsError,
    SystemMessage,
    UserMessage,
)
from any_llm_client.main import AnyLLMConfig, get_client
from any_llm_client.retry import RequestRetryConfig


__all__ = [
    "AnyLLMConfig",
    "AssistantMessage",
    "LLMClient",
    "LLMConfig",
    "LLMError",
    "Message",
    "MessageRole",
    "MockLLMClient",
    "MockLLMConfig",
    "OpenAIClient",
    "OpenAIConfig",
    "OutOfTokensOrSymbolsError",
    "RequestRetryConfig",
    "SystemMessage",
    "UserMessage",
    "YandexGPTClient",
    "YandexGPTConfig",
    "get_client",
]



import asyncio

import any_llm_client


config = any_llm_client.OpenAIConfig(url="http://127.0.0.1:11434/v1/chat/completions", model_name="qwen2.5-coder:1.5b")


async def main() -> None:
    async with any_llm_client.get_client(config) as client:
        print(await client.request_llm_message("Кек, чо как вообще на нарах?"))


asyncio.run(main())



import asyncio

import any_llm_client


config = any_llm_client.OpenAIConfig(url="http://127.0.0.1:11434/v1/chat/completions", model_name="qwen2.5-coder:1.5b")


async def main() -> None:
    async with (
        any_llm_client.get_client(config) as client,
        client.stream_llm_message_chunks(
            messages=[
                any_llm_client.SystemMessage("Ты — опытный ассистент"),
                any_llm_client.UserMessage("Кек, чо как вообще на нарах?"),
            ],
            temperature=1.0,
        ) as message_chunks,
    ):
        async for chunk in message_chunks:
            print(chunk, end="", flush=True)


asyncio.run(main())



import contextlib  # noqa: A005
import dataclasses
import typing

import httpx
import stamina

from any_llm_client.retry import RequestRetryConfig


DEFAULT_HTTP_TIMEOUT: typing.Final = httpx.Timeout(None, connect=5.0)


def get_http_client_from_kwargs(kwargs: dict[str, typing.Any]) -> httpx.AsyncClient:
    kwargs_with_defaults: typing.Final = kwargs.copy()
    kwargs_with_defaults.setdefault("timeout", DEFAULT_HTTP_TIMEOUT)
    return httpx.AsyncClient(**kwargs_with_defaults)


async def make_http_request(
    *,
    httpx_client: httpx.AsyncClient,
    request_retry: RequestRetryConfig,
    build_request: typing.Callable[[], httpx.Request],
) -> httpx.Response:
    @stamina.retry(on=httpx.HTTPError, **dataclasses.asdict(request_retry))
    async def make_request_with_retries() -> httpx.Response:
        response: typing.Final = await httpx_client.send(build_request())
        response.raise_for_status()
        return response

    return await make_request_with_retries()


@contextlib.asynccontextmanager
async def make_streaming_http_request(
    *,
    httpx_client: httpx.AsyncClient,
    request_retry: RequestRetryConfig,
    build_request: typing.Callable[[], httpx.Request],
) -> typing.AsyncIterator[httpx.Response]:
    @stamina.retry(on=httpx.HTTPError, **dataclasses.asdict(request_retry))
    async def make_request_with_retries() -> httpx.Response:
        response: typing.Final = await httpx_client.send(build_request(), stream=True)
        response.raise_for_status()
        return response

    response: typing.Final = await make_request_with_retries()
    try:
        yield response
    finally:
        await response.aclose()



import copy
import typing

import httpx

from any_llm_client.http import DEFAULT_HTTP_TIMEOUT, get_http_client_from_kwargs


class TestGetHttpClientFromKwargs:
    def test_http_timeout_is_added(self) -> None:
        original_kwargs: typing.Final = {"mounts": {"http://": None}}
        passed_kwargs: typing.Final = copy.deepcopy(original_kwargs)

        client: typing.Final = get_http_client_from_kwargs(passed_kwargs)

        assert client.timeout == DEFAULT_HTTP_TIMEOUT
        assert original_kwargs == passed_kwargs

    def test_http_timeout_is_not_modified_if_set(self) -> None:
        timeout: typing.Final = httpx.Timeout(7, connect=5, read=3)
        original_kwargs: typing.Final = {"mounts": {"http://": None}, "timeout": timeout}
        passed_kwargs: typing.Final = copy.deepcopy(original_kwargs)

        client: typing.Final = get_http_client_from_kwargs(passed_kwargs)

        assert client.timeout == timeout
        assert original_kwargs == passed_kwargs



import contextlib
import dataclasses
import types
import typing

import pydantic
import typing_extensions

from any_llm_client.core import LLMClient, LLMConfig, Message


class MockLLMConfig(LLMConfig):
    response_message: str = ""
    stream_messages: list[str] = pydantic.Field([])
    api_type: typing.Literal["mock"] = "mock"


@dataclasses.dataclass(slots=True)
class MockLLMClient(LLMClient):
    config: MockLLMConfig

    async def request_llm_message(
        self,
        messages: str | list[Message],  # noqa: ARG002
        *,
        temperature: float = 0.2,  # noqa: ARG002
        extra: dict[str, typing.Any] | None = None,  # noqa: ARG002
    ) -> str:
        return self.config.response_message

    async def _iter_config_stream_messages(self) -> typing.AsyncIterable[str]:
        for one_message in self.config.stream_messages:
            yield one_message

    @contextlib.asynccontextmanager
    async def stream_llm_message_chunks(
        self,
        messages: str | list[Message],  # noqa: ARG002
        *,
        temperature: float = 0.2,  # noqa: ARG002
        extra: dict[str, typing.Any] | None = None,  # noqa: ARG002
    ) -> typing.AsyncIterator[typing.AsyncIterable[str]]:
        yield self._iter_config_stream_messages()

    async def __aenter__(self) -> typing_extensions.Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None: ...
